#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This script helps to design a first-order passive RC low-pass filter that 
satisfies two requirements:
    
- The output noise voltage should be below a specified value
- The n-bit settling time should below a specified value

The script calculates the design limits (show-stopper values) for R and C; the
designer selects valid values for R and C.
"""
import SLiCAP as sl               # SLiCAP module
import sympy as sp                # Sympy module
from scipy.optimize import fsolve # Numeric solver from the Scipy module
import webbrowser                 # Module to address your default browser

# Put the specifications on top, so you can easily change them
v_n_max = 10e-6
tau_n_s = 1e-6
n_s     = 8

# Define symbols
R, C, t  = sp.symbols('R, C, t')

# Create/update the project directory structure and initialize SLiCAP
sl.initProject("RC network")

# Define and save the specifications
specs = []

specs.append(sl.specItem("v_n_max",
                         description = "Maximum RMS output voltage noise",
                         value       = v_n_max,
                         units       = "V",
                         specType    = "performance"))

specs.append(sl.specItem("tau_n_s",
                         description = "Maximum n-bit settling time",
                         value       = tau_n_s,
                         units       = "s",
                         specType    = "performance"))

specs.append(sl.specItem("n_s",
                         description = "Number of bits for $\\tau_{n_s}$",
                         value       = n_s,
                         units       = "",
                         specType    = "performance"))

cir = sl.makeCircuit("kicad/RCnetwork/RCnetwork.kicad_sch", imgWidth=300, 
                  description="RCnetwork.txt")

# Let us calculate the RMS output noise voltage
f_max = sp.Symbol("f_max", positive=True)

noiseResult = sl.doNoise(cir, pardefs='circuit')
# Sympy needs some help to calculate the integrated noise up to f=infinity
# you may know the answer: sqrt(k*T/C)
V_n_1       = sl.rmsNoise(noiseResult, 'onoise', 0, sp.oo)#f_max)

# Solve design equation for C_min, try symbolic, if it fails try numeric

C_min       = sp.solve(sl.float2rational(sp.simplify(V_n_1**2 - v_n_max**2)), sp.Symbol("C"))
if len(C_min) > 0:
    C_min = C_min[0]
else:
    print("Trying numeric solution")
    V_n_1_func = sp.lambdify(C, V_n_1**2 - v_n_max**2)
    C_min = fsolve(V_n_1_func, 1e-9)
    if len(C_min) > 0:
        C_min = C_min[0]
    else:
        print("Could not find a minimum capacitance C.")
        C_min = sp.oo

# Add to the specs
specs.append(sl.specItem("C_min",
                         description = "Minimum capacitance $C$",
                         value       = C_min,
                         units       = "F",
                         specType    = "design"))

# Let the designer select the value of C
C_sel = 0
while C_sel <= C_min:
    try:
        C_sel = float(input("Select the capacitance C > {}.\n>>> ".format(str(sp.N(C_min, 3)))))
    except ValueError:
        C_sel = 0

# Add selected value to the specs
specs.append(sl.specItem("C_sel",
                         description = "Selected capacitance $C$",
                         value       = C_sel,
                         units       = "F",
                         specType    = "design"))        

# Calculate R_max from the settling time and C_min
# We will use a two-step approach, first express the settling time in R and C; 
# later, we will use this expression for evaluatuion after selection of R and C
stp_s = sl.doStep(cir).stepResp
err_s = 1 - stp_s
try:
    # Try symbolic analysis
    tau_s = sp.solve(err_s - 2**(-n_s), t)[0]
    # Now substitute C and solve R; if it fails try numeric
    R_max = sp.solve(tau_s.subs(C, C_sel) - tau_n_s, R)[0]
except IndexError:
    # Except numeric analysis
    cir.defPar("C", C_sel)
    stp_n = sl.doStep(cir, pardefs="circuit").stepResp
    err_n = sp.lambdify(R, (1 - stp_n).subs(t, tau_n_s)- 2**(-n_s))
    R_max = fsolve(err_n, tau_n_s/C_sel)
    if len(R_max) == 0:
        print("Could not find maximum resistance")
        R_max = 0
    else:
        R_max = R_max[0]

# Add to the specs
specs.append(sl.specItem("R_max",
                         description = "Maximum resistance $R$",
                         value       = R_max,
                         units       = "Omega",
                         specType    = "design"))

# Let the designer select the value of R
R_sel = sp.oo
while R_sel >= R_max:
    try:
        R_sel = float(input("Select the resistance R < {}.\n>>> ".format(str(sp.N(R_max, 3)))))
    except ValueError:
        R_sel = sp.oo

# Evaluate the settling time obtained:
try:
    # Try symbolic evaluation from tau_s
    tau_s = sp.N(tau_s.subs({C: C_sel, R: R_sel}))
except IndexError:
    # Except solving it numerically
    f_err  = sp.lambdify(t, err_s.subs({R: R_sel, C: C_sel})- 2**(-n_s))
    try:
        tau_s  = fsolve(f_err, 0)[0]
    except IndexError:
        print("Error: could not evaluate tau_s.")
        tau_s = None
        
# Add to the specs
specs.append(sl.specItem("R_sel",
                         description = "Selected resistance $R$",
                         value       = R_sel,
                         units       = "Omega",
                         specType    = "design"))   

# Add some achieved specifications"
specs.append(sl.specItem("B",
                         description = "Achieved -3dB Bandwidth",
                         value       = sp.N(1/2/sp.pi/R_sel/C_sel),
                         units       = "Hz",
                         specType    = "design"))

specs.append(sl.specItem("B_n",
                         description = "Achieved noise bandwidth",
                         value       = sp.N(1/4/R_sel/C_sel),
                         units       = "Hz",
                         specType    = "design"))

specs.append(sl.specItem("V_n",
                         description = "Achieved RMS output noise",
                         value       = sp.N(V_n_1.subs({C: C_sel, R: R_sel})),
                         units       = "V",
                         specType    = "design"))

specs.append(sl.specItem("tau_s",
                         description = "Achieved $n_s$ bit settling time",
                         value       = tau_s,
                         units       = "s",
                         specType    = "design"))

sl.specs2csv(specs, "specs.csv")

# HTML Report #################################################################

sl.htmlPage("Initial specifications")
sl.specs2html(specs, types=["performance"])

sl.htmlPage("Design equations")
sl.head2html("Circuit diagram")
sl.img2html("RCnetwork.svg", 300)
sl.head2html("Initial specifications")
sl.specs2html(specs, types=["performance"])
sl.head2html("Output voltage noise")
sl.text2html("Spectral density of the output voltage noise")
sl.eqn2html("S_vno", noiseResult.onoise, units="V^2/Hz")
sl.text2html("RMS output noise voltage")
sl.eqn2html("V_no", V_n_1, units="V")
sl.head2html("Minimum capacitance")
sl.eqn2html("C_min", C_min)
sl.head2html("Selected capacitance")
sl.eqn2html("C_sel", C_sel)
sl.head2html("Unit step response")
sl.eqn2html("a(t)", stp_s)
sl.head2html("Error versus time")
sl.eqn2html("epsilon(t)", err_s)
sl.head2html("Settling error to $n_s$ bits")
sl.eqn2html("tau_s", tau_s)
sl.head2html("Maximum resistance")
sl.eqn2html("R_max", R_max)
sl.head2html("Selected resistance")
sl.eqn2html("R_sel", R_sel)

sl.htmlPage("Final specifications")
sl.specs2html(specs)

# Show the report
webbrowser.open_new(sl.ini.html_path + 'index.html')